package de.ims.cl.dv;

import java.util.*;
import java.io.*;

import de.ims.cl.dv.datastructures.*;
import de.ims.cl.dv.parsers.*;
import de.ims.cl.dv.util.*;

/** <b>executable</b> ExtractBasisElements extracts a list of
    BasisElements (equivalence classes of Paths). In addition, it
    counts BEs, target words and the total number of paths. For the
    command line parameters, see the help (just call
    bin/ExtractBasisElements.sh).
**/

public class ExtractBasisElements {
    Encoding e;
    Corpus corpus;
    Logfile log;
    double[][] totalcounts;
    ContextSpec[] cs;
    int maxsize; // maximal number of BasisElements
    double ratio; // ratio of infrequent basisElements to be cutoff every time maxSize is reached
    FrequencyList[][] targets; // targets and their frequencies
    String[][] targetcountsfile,becountsfile,totalcountsfile;
    String[] valuation;
    FrequencyList[][] basisElements; // basisElements and their frequencies
    String path;
    Parameters p;
    
    boolean isDebug = false;
    
    String validPOS;
    
    static String[] allCSFiles = Parameters.allCSFiles;
    static String[] allValuations = Parameters.allValuations;
    // TODO: do i need the posmap option here?
    static String[] basisOptions = {"corpus","spec","log","maxsize","cutoff","targets","targetdir","help","pos","basisMapping", "ctype", "posmap", "relmap"}; // recognized command line options
    
    
    private void showHelp() {
    
    System.err.println("ExtractBasisElements is part of the DependencyVectors package.");
    System.err.println("");
    System.err.println("ExtractBasisElements traverses a parsed corpus and extracts several statistics. These are: \n\t(1) the set of the n most frequent basis elements according to the given parametrisation,\n\t(2) the frequency of target words in the corpus and \n\t(3) the total number of basis elements according to the parametrisation. \n");
    System.err.println("");
    System.err.println("The results are mainly influenced by the path value and context specification function chosen.\nExtractBasisElements supports the concurrent extraction of multiple lists for different \nparametrisations.To use this feature, do not specify the parameter(s) you want to vary."); 
    System.err.println("The resulting lists are written to targetdir/<contextspec>_<pathvalue>.bes, .targets and .total\n");
    System.err.println("");
    System.err.println("ExtractBasisElements supports the following command line parameters:");
    System.err.println("\t--corpus <name>: Location of corpus [default: System.in]");
    System.err.println("\t--ctype <name>: Type of corpus [minipar / candc / freeling]");
    System.err.println("\t--cutoff <int>: ratio of basis elements to be deleted in each purge [default: 0.3]");
    System.err.println("\t--log <file>: destination for the log file [default: log.txt]");
    System.err.println("\t--maxsize <int>: desired number of basis elements [default: 10000]");
    System.err.println("\t--spec <file>:  file containing context specification (cspec file format) [default: all]");
    System.err.println("\t\tCurrently known specs: "+Parameters.prettyPrintSpecs());
    System.err.println("\t--targets <file>: file containing target words (word list file format)");
    System.err.println("\t--targetdir <file>: destination directory for the list files.");
    System.err.println("\t"+Parameters.prettyPrintVals()+": Valid path value functions [default: all]");
    System.err.println("");
    System.err.println("Example call: zcat bnc.parsetrees.gz | bin/ExtractBasisElements.sh --targets targetsfile\n              --targetdir data/bnc/ --plain --spec minimal");
    System.err.println("");
    System.exit(0);
    }    

    /** Evaluates command line parameters, prepares output directories, initialises data structures, etc. **/

    private ExtractBasisElements(DVCommandLineMap clm) throws IOException {
        
        if (clm.exists("help")) {
            showHelp();
        }

        // prepare output directory 
    
        path = clm.determineTargetDirectory();
        if (path == null) {
            System.err.println("\"--targetdir\" command line argument missing. \n ");
            showHelp();
        }

        // create logfile;
        log = clm.createLogfile(path);
        System.out.println("Logfile created");
    
        // determine path valuation function(s)
    
        valuation = clm.determineValueFunctions(allValuations,log);
        System.out.println("Valuation OK");
    
        // initialise context specification(s)

        cs = clm.determineContextSpecs(allCSFiles,log);
        System.out.println("determineContextSpecs OK");
    
        // init corpus

        validPOS = clm.validPOS();
        System.out.println("validPOS OK");
    
        //  System.err.println("Known POS: "+validPOS);
    
        corpus = clm.determineCorpus(log);
        System.out.println("determineCorpus OK");
    
        // intialise names of output files

        targetcountsfile = new String[cs.length][valuation.length];
        becountsfile = new String[cs.length][valuation.length];
        totalcountsfile = new String[cs.length][valuation.length];
        totalcounts = new double[cs.length][valuation.length];
    
        for (int i=0; i<cs.length; i++) {
            for (int j=0; j<valuation.length; j++){
            String affix = DVIOHelpers.affix(cs[i].name(),valuation[j]);
            targetcountsfile[i][j] = path + affix +".targets";
            becountsfile[i][j] =  path + affix + ".bes";
            totalcountsfile[i][j] = path + affix + ".total";
            DVIOHelpers.testLocation(becountsfile[i][j]);
            DVIOHelpers.testLocation(targetcountsfile[i][j]);
            DVIOHelpers.testLocation(totalcountsfile[i][j]);
            }
        }
    
        // initialise targets and basis elements data structures

        targets = new FrequencyList[cs.length][valuation.length];
        basisElements = new FrequencyList[cs.length][valuation.length];
    
        if (clm.exists("targets")) {
            log.println("Reading target words from file "+clm.get("targets")+".");
            
            
            if (!clm.exists("ctype")) {
                System.err.println("Corpus type not specified.");
                showHelp();
            }
            if (clm.get("ctype").equals("minipar")) {
                e = new MiniparEncoding();
            }
            else if (clm.get("ctype").equals("candc")) {
                e = new CandCEncoding();               
            }
            else if (clm.get("ctype").equals("freeling")) {
            	e = new FreelingEncoding();            	
            }
            else {
            	System.err.println("Unknown corpus type.");
            	showHelp();
            }
            
            for (int i=0; i<cs.length; i++) {
                for (int j=0; j<valuation.length; j++){
                    targets[i][j] =  new FrequencyList(clm.get("targets"), log,false,e);
                    basisElements[i][j] = new FrequencyList(log);
                }
            }
        } else {
            System.err.println("\"--targets\" command line argument missing. Exiting. \n ");
            log.println("\"--targets\" command line argument missing. Exiting. \n ");
            showHelp();
        }
    
        if (clm.exists("maxsize")) {
            try {
            maxsize = Integer.parseInt(clm.get("maxsize"));
            log.println("Maxsize is "+clm.get("maxsize")+".");
            } catch (NumberFormatException e) {
            log.println("DEFAULT: Illegal maxsize. Reverting to 10000.");
            maxsize = 10000;
            }
        } else { 
            log.println("DEFAULT: Maxsize is 10000 (default).");
            maxsize = 10000;
        }
    
        if (clm.exists("cutoff")) {
            try {
            ratio = Double.parseDouble(clm.get("cutoff"));
            log.println("Cutoff ratio is "+clm.get("cutoff")+".");
            } catch (NumberFormatException e) {
            log.println("DEFAULT: Illegal ratio. Reverting to 0.3");
            ratio = 0.3;
            }
        } else { 
            log.println("DEFAULT:  Ratio is 0.5.");
            ratio = 0.5;
        }
        if ((ratio<0) || (ratio > 100)) {
            log.println("DEFAULT: Ratio not in permitted interval: Reverting to 0.5.");
            ratio = 0.5;
        }
    
      p = new Parameters(clm.getBasisMapping(log),log);
    
 }

    /** read a corpus **/
    
    private void readCorpus(Corpus corpus) throws IOException {
    
    // read each paragraph of the corpus, construct a graph, and count equivalence classes.

    int sentcounter = 0; 
    
    while (corpus.hasNext()) {
            Graph parseGraph = corpus.next();         
            if (parseGraph == null)
            	break;        
            if (parseGraph.isEmpty()) { // an error occurred during parsing
                // DEBUG
                // System.out.println("empty parse graph");
                continue;
            }
            
            
//            System.out.println(parseGraph);
            
            
            for (int i=0; i<cs.length; i++) {
                for (int j=0; j<valuation.length; j++){     
                    basisElements[i][j].purge(maxsize,ratio);
                }
            }

            Set<Integer> ws = parseGraph.getAllNodeIndices();     
        
            // HIER
            sentcounter++;
            //      System.err.println("Sent "+sentcounter);
        
            //      if (sentcounter < 3)
            //      continue;
            //      if (sentcounter > 3)
            //      System.exit(0);

            for (Integer index: ws) { // for all nodes in the parse graph
                String word = parseGraph.getNode(index).getWord();
                String pos = parseGraph.getNode(index).getPOS();
                // System.err.println("Found POS "+pos);
                boolean isTarget = (targets[0][0].containsKey(word)); // all target lists are identical
                for (int i=0; i<cs.length; i++) {                     // NEW: ITERATE OVER ALL CSs
                    Context c = new Context(parseGraph,cs[i],index);   //  contruct context
                  //  System.err.println("Context size "+c.size());

                    if (isTarget && isDebug) {
                        System.err.println("Word "+word+"\tcontext size "+c.size());            
                        System.err.println("POS: " + pos);
                    }
                  
                    for (Path contextPath: c) {
                        String thisBasisElement = e.encode(p.basisMapping(contextPath)); // compute unique path representation
                        for (int j=0; j<valuation.length; j++) {                
                            double value = p.pathValue(contextPath,valuation[j]);
                            totalcounts[i][j] = totalcounts[i][j] + value;
                            Double valueD = new Double(value);
                            if (isTarget && (validPOS == null || pos.equals(validPOS))) { 
                            targets[i][j].inc(word, valueD); }
                            
                            if (thisBasisElement.matches("\\s*")) {
                                System.out.println("EMPTY BE");
                                System.out.println("Context Path: " + contextPath.toString());;
                                System.out.println("GRAPH: " + parseGraph.toString());
                                System.exit(0);
                            }
                            
                            if (basisElements[i][j].containsKey(thisBasisElement)) {
                            // System.out.println("Incrementing "+uniqRep+" by "+value);
                            basisElements[i][j].inc(thisBasisElement, valueD); }
                            else {
                            basisElements[i][j].put(thisBasisElement, valueD); 
                            }       
                        }
                    }
                }
            }
        }
    }

    /** write to file **/

    private void writeFiles() throws IOException {  
    for (int i=0; i<cs.length; i++) {
        for (int j=0; j<valuation.length; j++){     
        basisElements[i][j].purge(maxsize,1.0);
        DVIOHelpers.writeToFile(becountsfile[i][j],basisElements[i][j].toStringByFrequency());
        DVIOHelpers.writeToFile(targetcountsfile[i][j],targets[i][j].toStringByFrequency());
        DVIOHelpers.writeToFile(totalcountsfile[i][j],new Double(totalcounts[i][j]).toString());
        }
    }
    }
    
    
    /** This function is called when running the program with <i>java
     * ExtractBasisElements \< arguments \> </i>. The program relies on
     * receiving the right command line arguments and will exit if
     * they are not supplied. See help output and documentation for
     * details about arguments.
     **/
    static public void main(String[] args){
    
    ArrayList<String> recognisedOptions = new ArrayList<String>();
    recognisedOptions.addAll(Arrays.asList(basisOptions));
    recognisedOptions.addAll(Arrays.asList(Parameters.allValuations));
    
    DVCommandLineMap arguments = new DVCommandLineMap(args,recognisedOptions);
    try {
        // DEBUG
        // System.out.println("INITIALIZING ARGUMENTS FOR EXTRACTBASISELEMENTS");
        ExtractBasisElements ebe = new ExtractBasisElements(arguments);
        // DEBUG
        // System.out.println("READING CORPUS FOR EXTRACTBASISELEMENTS");
        ebe.readCorpus(ebe.corpus);
        // DEBUG
        // System.out.println("WRITING FILES FOR EXTRACTBASISELEMENTS");
        ebe.writeFiles();
        ebe.log.close();         // side effect: write current time into log
    } catch (IOException e) {
        System.err.println("[ExtractBasisElements] IO Error: "+e);      
    }
    }
}
