package de.ims.cl.dv;

import java.util.*;
import java.io.*;

import de.ims.cl.dv.datastructures.*;
import de.ims.cl.dv.parsers.*;
import de.ims.cl.dv.util.*;
import de.ims.cl.dv.space.*;

/** <b>executable</b> ExtractIncrementalSpace constructs a SemanticSpace for a given set of target words, and a given set of BasisElements
    to use as dimensions. The lists of Basis Elements can be produced by the ExtractBasisElements class.
 **/
   
public class ExtractIncrementalSpace {

    SemanticSpace[][] semanticSpaces;
    String[][] spaceFiles;
    Corpus corpus;
    ContextSpec[] cs;
    Logfile log;
    String[] valuation;
    FrequencyList[][] targets;
    FrequencyList[][] basisElements;
    long every;
    String path;
    boolean complainAboutFrequencies; 
    static String[] basisOptions = {"corpus","spec","log","help","every","targetdir","small","nofrequencies","basisMapping","pos", "ctype"};
    double[][] totalcounts;
    Parameters p;
    
    static String[] allValuations = Parameters.allValuations;
    static String[] allCSFiles = Parameters.allCSFiles;    
    
    private void showHelp() {
	System.err.println("ExtractBasisElements is part of the DependencyVectors package.\n");
	System.err.println("");
	System.err.println("ExtractIncrementalSpace traverses a parsed corpus and constructs vectors for target words.\nThe dimensions are the given basis elements.");
	System.err.println("The results are mainly influenced by the path value and context specification function chosen.\nExtractIncrementalSpaces supports the concurrent extraction of multiple spaces for different\nparametrisations. To use this feature, do not specify the parameter(s) you want to very."); 
	System.err.println("The basis elements are read from targetdir/<contextspec>_<pathvalue>.bes");
	System.err.println("The target words are read from targetdir/<contextspec>_<pathvalue>.targets");
	System.err.println("The resulting semantic spaces are written to targetdir/<contextspec>_<pathvalue>.vectors\n");
	System.err.println("");
	System.err.println("ExtractIncrementalSpaces supports the following command line parameters:");
	System.err.println("\t--corpus <file>: MINIPAR-parsed corpus file [default: System.in]");
	System.err.println("\t--every <num>: Incremental mode: Write semantic space every <num> corpus bytes\n\t               to the file targetdir/<contextSpec_valuation.space.<num>.");
	System.err.println("\t--log <file>: destination for the log file [default: log.txt]");
	System.err.println("\t--small: Uses Semantic Space class with small memory footprint (recommended for concurrent extracting of multiple spaces, but possibly slower)");
	System.err.println("\t--spec <file>:  file containing context specification (cspec file format) [default: all]");
	System.err.println("\t\tCurrently known specs: "+Parameters.prettyPrintSpecs());

	System.err.println("\t--targetdir <file>: directory in which the basis elements live and  the semantic space is to be written.");
	System.err.println("\t"+Parameters.prettyPrintVals()+": Valid path value functions [default: all]");

        System.err.println("\t--nofrequencies: run also with target and basis elements files without frequencies (no log-likelihood transformation possible)");												       
        System.err.println("");
	System.err.println("Example call: zcat bnc.parsetrees.gz | bin/ExtractIncrementalSpace.sh --targetdir targets\n              --plain --spec contextspec_minimal.txt");
	System.exit(0);
    }    

 
    /** initialise data structured **/
 
    private ExtractIncrementalSpace(DVCommandLineMap clm) throws IOException {	

	if (clm.exists("help")) {
	    showHelp();
	}

	// prepare output directory	
	
	path = clm.determineTargetDirectory();
	if (path == null) {
	    System.err.println("\"--targetdir\" command line argument missing. \n ");
	    showHelp();
	}

	// create logfile;

	log = clm.createLogfile(path);
	
	// determine path valuation function(s)
	
	valuation = clm.determineValueFunctions(allValuations,log);
	
	// initialise context specification(s)

	cs = clm.determineContextSpecs(allCSFiles,log);
	
	// init corpus

	corpus = clm.determineCorpus(log);

	// init every

	if (clm.exists("every")) {
	    log.println("Output every "+clm.get("every")+" corpus bytes.");
	    every = Integer.parseInt(clm.get("every")); }
	else {
	    every = 0;
	}
	
	// initialise semantic spaces, and test whether we can write semantic spaces

	targets = new FrequencyList[cs.length][valuation.length];
	basisElements = new FrequencyList[cs.length][valuation.length];

	if (clm.exists("small")) {
	    log.println("Using HashMap-based spaces (--small)");
	    semanticSpaces = new SemanticSpaceHash[cs.length][valuation.length];
	} else {
	    log.println("Using array-based spaces");
	    semanticSpaces = new SemanticSpaceArray[cs.length][valuation.length];
	}
	
	complainAboutFrequencies = (!(clm.exists("nofrequencies")));
	
	//	spaceFiles = new String[cs.length][valuation.length]; 
	
	for (int i=0; i<cs.length; i++) {
	    for (int j=0; j<valuation.length; j++){		
		String affix = DVIOHelpers.affix(cs[i].name(),valuation[j]);

		targets[i][j] = new FrequencyList(path+affix+".targets",log,complainAboutFrequencies);
		targets[i][j].resetFrequencies();

		basisElements[i][j] = new FrequencyList(path+affix+".bes",log,complainAboutFrequencies);
		basisElements[i][j].resetFrequencies();
		
		if (clm.exists("small")) {
		    semanticSpaces[i][j] = new SemanticSpaceHash(targets[i][j],basisElements[i][j],log);
		} else {
		    semanticSpaces[i][j] = new SemanticSpaceArray(targets[i][j],basisElements[i][j],log); }
		//		spaceFiles[i][j] = path+affix+".vectors";
		//		DVIOHelpers.testLocation(spaceFiles[i][j]);
	    }
	}
  p = new Parameters(clm.getBasisMapping(log),log);
	log.println("Initialisation finished.");
    }


    /** traverse corpus and collect counts **/
    
    private void readCorpus(Corpus corpus) throws IOException {
	long no_every = 0; // how often we have dumped into a file

	totalcounts = new double[cs.length][valuation.length];
		
       	while (corpus.hasNext()) {
	    Graph parseGraph = corpus.next();	  
	    if (parseGraph.isEmpty()) { // an error occurred during parsing
		continue;}
	    Set<Integer> ws = parseGraph.getAllNodeIndices();
	    Iterator<Integer> wsit = ws.iterator();
	    while (wsit.hasNext()) { // for all nodes in the parse graph
		Integer index = (Integer) wsit.next();
		String targetWord = parseGraph.getNode(index).getWord();		

		for (int i=0; i<cs.length; i++) {                     // NEW: ITERATE OVER ALL CSs
		    Context c = new Context(parseGraph,cs[i],index);   //  contruct context
		    Iterator<Path> pathit = c.iterator();  // for all paths in the context
		    while (pathit.hasNext()) {
			Path contextPath = (Path) pathit.next();
			String thisBasisElement = p.basisMapping(contextPath); // compute unique path representation
			
			for (int j=0; j<valuation.length; j++) {	
			    boolean isBE = (basisElements[i][j].containsKey(thisBasisElement));
			    boolean isTarget = (targets[i][j].containsKey(targetWord)); 
			    double value = p.pathValue(contextPath,valuation[j]);
			    
			    totalcounts[i][j] = totalcounts[i][j] + value;
			    if (isTarget) {targets[i][j].inc(targetWord, new Double(value)); }
			    if (isBE) { basisElements[i][j].inc(thisBasisElement, new Double(value)); }
			    if (isTarget && isBE) {
				semanticSpaces[i][j].increaseCount(targetWord,thisBasisElement,value);
			    }
			}
		    }
		}
	    }
	    if ((every != 0) && ((corpus.charcounter() / every) > no_every)) { // write incremental space
		no_every = corpus.charcounter() / every;
		for (int i=0; i<cs.length; i++) {
		    for (int j=0; j < valuation.length; j++) {			
			DVIOHelpers.writeToFile(path+DVIOHelpers.affix(cs[i].name(),valuation[j])+"_"+no_every+".vectors",semanticSpaces[i][j].toString());
			DVIOHelpers.writeToFile(path+DVIOHelpers.affix(cs[i].name(),valuation[j])+"_"+no_every+".bes",basisElements[i][j].toStringByFrequency());
			DVIOHelpers.writeToFile(path+DVIOHelpers.affix(cs[i].name(),valuation[j])+"_"+no_every+".targets",targets[i][j].toStringByFrequency());
			DVIOHelpers.writeToFile(path+DVIOHelpers.affix(cs[i].name(),valuation[j])+"_"+no_every+".total",new Double(totalcounts[i][j]).toString());
		    }
		}
	    }
	}
    }
    
    /** write semantic spaces to files **/

    private void writeFiles() throws IOException{	
	log.println("Finished reading. Now writing to files.");	
	for (int i=0; i<cs.length; i++) {                     // iterate over all context specs
	    for (int j=0; j < valuation.length; j++) {			
		DVIOHelpers.writeToFile(path+DVIOHelpers.affix(cs[i].name(),valuation[j])+".vectors",semanticSpaces[i][j].toString());
		DVIOHelpers.writeToFile(path+DVIOHelpers.affix(cs[i].name(),valuation[j])+".bes",basisElements[i][j].toStringByFrequency());
		DVIOHelpers.writeToFile(path+DVIOHelpers.affix(cs[i].name(),valuation[j])+".targets",targets[i][j].toStringByFrequency());
		DVIOHelpers.writeToFile(path+DVIOHelpers.affix(cs[i].name(),valuation[j])+".total",new Double(totalcounts[i][j]).toString());		
	    }
	}		
    }
    
    /** This function is called when running the program with <i>java
     * ExtractBasisElements /< arguments />. The program relies on
     * receiving the right command line arguments and will exit if
     * they are not supplied. See help output and documentation for
     * details about arguments. </i>
     **/
    public static void main(String[] argarray) throws IOException {
	try {
	    ArrayList<String> recognisedOptions = new ArrayList<String>();
	    recognisedOptions.addAll(Arrays.asList(basisOptions));
	    recognisedOptions.addAll(Arrays.asList(Parameters.allValuations));
	    DVCommandLineMap arguments = new DVCommandLineMap(argarray,recognisedOptions);
	    ExtractIncrementalSpace es = new ExtractIncrementalSpace(arguments);
	    es.readCorpus(es.corpus);
	    es.writeFiles();
	    es.log.close(); // close logfile (side effect: write current time into logfile)	
	} catch (IOException e) {
	    System.err.println("[ExtractIncrementalSpace] IO Error: ]"+e);
	}
	
    }
}
