/**
   class ContextSpec (and class BadFilenameException)

   Copyright (C) 2002-4 Sebastian Pado
   
   This file is part of the DependencyVectors v2 distribution.
   
   DependencyVectors is free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later
   version.
   
   DependencyVectors is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public
   License along with DependencyVectors; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA

**/

package de.ims.cl.dv.datastructures;


import java.util.*;
import java.io.*;
import java.util.regex.*;

import de.ims.cl.dv.parsers.*;

/** This exception is thrown if the filename of a context specification cannot be analysed.
    The current pattern is /.*_(\w+)/ where the name is the bit after the underscore.
**/ 

class BadFilenameException extends RuntimeException {
    private static final long serialVersionUID = 1;
    BadFilenameException(String errormsg) {
    super(errormsg);
    }
}

/** ContextSpec models the context specification function. In the current implementation, it is only 
    a set of path specifications.
**/


public class ContextSpec {

    Set<Path> contextSpec;

    String name;
    
    /** returns an iterator over all path specifications
     **/
    
    public Iterator<Path> iterator() {
	return contextSpec.iterator();
    }

    /** returns the name of the context specification, determined from the initialiasation filename. 
	The current required pattern for the filename is /.*_(\w+)/ where the name is the bit after the underscore.
    **/
    
    public String name() {
	return name;
    }
    
    /** A ContextSpec is initialized with the filename of a file
	containing path specifications in textual form. See the
	(external) documentation of DependencyVectors for a
	description of the format for path specification files.
    **/
    
    public ContextSpec(String filename) {
	contextSpec = new HashSet<Path>();
	try {
	    BufferedReader contspecs = new BufferedReader(new FileReader(filename)); // open specification file
	    String line;
	    while ((line = contspecs.readLine()) != null) { // for every line:
		if (line.equals("")) { continue; } // ignore empty lines
		if (line.startsWith("#")) { continue; } // ignore comments
		line = line.split("\t")[0];
		Path myPathSpec = new Path(line,new MiniparEncoding());
		this.contextSpec.add(myPathSpec);
	    }
	    Pattern r = Pattern.compile("\\_(\\w+)");
	    // Matcher m = r.matcher(new File(filename).getName());
	    Matcher m = r.matcher(filename);
	    if (m.find()) {
		name = m.group();
		name = name.substring(1,name.length());
	    } else {
		throw new BadFilenameException("Could not parse name of ContextSpec file: "+filename);
	    }
	}
	catch (IOException e) { 
	    System.out.println("Error: "+e);
	    System.exit(0);
	}
	catch (NullPointerException e) {
	    System.out.println("Could not parse context specification line "+e);
	    System.exit(0);
	}
    }        
}

