/**
   class FrequencyList (and class SortInvFrequency)

   Copyright (C) 2002-4 Sebastian Pado
   
   This file is part of the DependencyVectors v2 distribution.
   
   DependencyVectors is free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later
   version.
   
   DependencyVectors is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public
   License along with DependencyVectors; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA

**/

package de.ims.cl.dv.datastructures;

import java.util.*;
import java.io.*;

import de.ims.cl.dv.parsers.Encoding;
import de.ims.cl.dv.util.Logfile;

// this is a helper class to sort a Map according to the inverse of its values.
// [DEBUGGED VERSION Mar 27 2011]

class SortInvFrequency implements Comparator<Map.Entry<String,Double>> {
    
    public int compare(Map.Entry<String,Double> o1, Map.Entry<String,Double> o2) {
	Double freq1 = o1.getValue();
	Double freq2 = o2.getValue();
	int freqComparison = -(freq1.compareTo(freq2));
	if (freqComparison == 0) {
//		System.err.println("same freq for "+o1.getKey()+" and "+o2.getKey()+". Result is "+o1.getKey().compareTo(o2.getKey()));
		return -(o1.getKey().compareTo(o2.getKey()));
	} else {
		return freqComparison;
		}
    }
}

/** This class models a hash map whose keys are Strings, and values are Doubles to keep track of frequency counts.  
    The hash table can be kept at constant size by removing the most infrequent entries. 
**/

public class FrequencyList extends HashMap<String,Double> {
    private static final long serialVersionUID = 1;
    Logfile log;

    /** Increase the frequency count for o by increment. **/
    
    public void inc(String o, Double increment) {
	put(o, new Double(((Double) get(o)).doubleValue() + increment.doubleValue()));
    }

    /** Reduces the FrequencyList to size maxsize if its size is
	larger than maxsize/ratio.  This works by removing infrequent
	entries until the FrequencyList is small enough. There is no
	guarantee that the FrequencyList is exactly of size maxsize
	after purge has been called (can be smaller). If the
	FrequencyList is smaller than maxsize/ratio, nothing happens.
    **/
    
    public void purge(int maxsize, double ratio) throws IOException {  // remove infrequent stuff
	int origSize = size(); 
	if (origSize > (maxsize/ratio)) {
	    List<Map.Entry<String, Double>> thisEntries = entriesByFrequency();    
	    Iterator<Map.Entry<String,Double>> it = thisEntries.iterator(); // find cutoff
	    int counter = 1;
	    Map.Entry<String,Double> currentEntry = it.next();
	    while (counter < maxsize) {
		counter++;
		currentEntry = it.next();
	    }
	    double co =  currentEntry.getValue();

	    List<String> toBeDeleted = new LinkedList<String>();
	    
	    for (String key: this.keySet()) {
		Double freq = (Double) get(key);
		if (freq.doubleValue() <= co) {
		    toBeDeleted.add(key);
		}
	    }
	    
	    for (String deleteKey: toBeDeleted) {
		this.remove(deleteKey);
	    }
	    log.println("Purged from size "+origSize+" to "+size());
 	}
    }

    /** Returns a list representation of the entries in the
     * FrequencyList, sorted according to inverse frequency (largest
     * first). List entries are of type Map.Entry **/
    
    public List<Map.Entry<String,Double>> entriesByFrequency() {  // elements of list are of type Map.Entry
	List<Map.Entry<String,Double>> l = new ArrayList<Map.Entry<String,Double>>(this.entrySet());
	Collections.sort(l, new SortInvFrequency());
	return l;
    }
    
    /** Returns a list representation of the entries in the
     * FrequencyList, sorted according to the keys. List entries are
     * of type Map.Entry. **/

    public List<Map.Entry<String, Double>> entriesNatural() {
	List<Map.Entry<String, Double>> l = new ArrayList<Map.Entry<String, Double>>((new TreeMap<String, Double>(this)).entrySet()); // this sorts in ascending key order
	return l;
    }

    private String privateToString(List<Map.Entry<String,Double>> list) {
    // use a StringBuffer for efficiency reasons
	StringBuffer r = new StringBuffer(list.size()*10);
	Iterator<Map.Entry<String,Double>> it = list.iterator();
	while (it.hasNext()) {
	    Map.Entry<String,Double> next = it.next();
	    Object item = next.getKey();
	    Object freq = next.getValue();
	    r.append(item);
	    r.append("\t");
	    r.append(freq);
	    r.append("\n");
	}
	return r.toString();
    }


    /** Returns a string representation of the FrequencyList, 
	sorted inversely according to the values (greatest frequency first). **/ 

    public String toStringByFrequency() {
	return privateToString(entriesByFrequency());	
    }

    /** Returns a string representation of the FrequencyList, sorted
     * according to keys. **/

    public String toStringNatural() {
	return privateToString(entriesNatural());
    }    

    /** Set all frequencies to zero **/
    
    public void resetFrequencies() {
	for (String key: this.keySet()) {
	    put(key,new Double(0));
	}
    }
    
    /** Initialise an empty FrequencyList.  **/

    public FrequencyList(Logfile log) {
	super();
	
	this.log = log;
    }
    
    /** Initialise the FrequencyList from a file with a frequency
	list. The format is recorded in the external documentation for
	DependencyVectors, and is produced by the toStringX methods.  **/
    
    public FrequencyList(String filename, Logfile log, boolean complainAboutMissingFrequency) throws IOException, FileNotFoundException {
	this(filename,log,complainAboutMissingFrequency, new Encoding());
	    }

    public FrequencyList(String filename, Logfile log, boolean complainAboutMissingFrequency, Encoding e) throws IOException, FileNotFoundException { // initialise from file
	super();
	
	this.log = log;

	boolean loggedWarning = false;
	
	String line;
	
	    BufferedReader targetFileReader = new BufferedReader(new FileReader(filename));
	    // System.err.println("Start reading from "+filename);
	    while ((line = targetFileReader.readLine()) != null) {
		StringTokenizer linet = new StringTokenizer(line,"\t"); 
		
		if (!(linet.hasMoreTokens())) continue; // allow for empty lines

		String target = linet.nextToken(); // cannot use trim since trim		
		Double frequency = new Double(0.0);
		// is there a frequency in the file?
		if (linet.hasMoreTokens()) {
		    frequency = Double.valueOf(linet.nextToken()); }
		else {
		    if (complainAboutMissingFrequency) {
			throw new IOException("No frequencies in "+filename+", line "+line);
		    } else {
			if (!(loggedWarning)) {
			    log.println("Warning: no frequencies in "+filename); 
			    loggedWarning = true;
			}
		    }
		}
				
		put(e.encode(target),frequency);
	    }
    }
    
}
				   