/**
   class Graph (and UnknownNodeException)

   Copyright (C) 2002-4 Sebastian Pado
   
   This file is part of the DependencyVectors v2 distribution.
   
   DependencyVectors is free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later
   version.
   
   DependencyVectors is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public
   License along with DependencyVectors; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA

**/

package de.ims.cl.dv.datastructures;

import java.util.*;

// is thrown if a node is requested that is unknown

class UnknownNodeException extends RuntimeException {
    private static final long serialVersionUID = 1;
    UnknownNodeException(String errormsg) {
        super(errormsg);
    }
}

/** This class models the graph that contains the dependency parsing
 * information for a sentence (one paragraph of input data).
 **/
    
public class Graph {
    
    
    
    public String toDot(String uniqID, Boolean directed){
        String result;
        Integer nodeid;
        Node   node;
        String lemma;
        String pos;
        
        Integer outid;
        Integer inid;
        @SuppressWarnings("unused")
		String edgelabel;
        String rel;
        
        Map<Integer,Integer> skip = new HashMap<Integer,Integer>();
        
        String edgeStyle;
        if (directed) {
            edgeStyle = " -> ";
        } else {
            edgeStyle = " -- ";
        }
        
        
        result = "\n# NODES\n";
        
        
        
        
        
        for (Map.Entry<Integer,Node> pair: nodes.entrySet()) {
            nodeid = pair.getKey();
            node   = pair.getValue();
            lemma  = node.getWord();
            pos    = node.getPOS();
            result += "\t" + uniqID+"_"+nodeid + " [label=\"" + lemma + " (" + pos +")" + "\" ]\n";
        }
        
        result += "\n# EDGES\n";
        
        for (Map.Entry<Integer,Map<Integer,Edge>> outNode: edges.entrySet()) {
            outid                     = outNode.getKey();
            Map<Integer,Edge> inNodes = outNode.getValue();
            for (Map.Entry<Integer,Edge> inNode: inNodes.entrySet()) {
                inid      = inNode.getKey();
                Edge edge = inNode.getValue();
                rel       = edge.getRelation();
                // add edge only if it (or reverse) haven't been added already
                if ((skip.containsKey(inid) && skip.get(inid).equals(outid)) || (skip.containsKey(outid) && skip.get(inid).equals(inid)))
                    continue;
                result += "\t" + uniqID+"_"+outid + edgeStyle + uniqID+"_"+inid + " [label = \"" + rel + "\"]\n";
                skip.put(outid, inid);
                
            }
        }
        return result;
        
    }
    
    public String toDot(String uniqID){
        return toDot(uniqID, false);
    }
    
    
    public String toDot(){
        return toDot("");
    }
    
    public String completeDot(Boolean directed) {
        return completeDot("", directed);
    }
    
    public String completeDot(String uniqID, Boolean directed){
        // Build a graphviz .dot representation of the graph
        @SuppressWarnings("unused")
		String gtype;
        String result;
        if (directed) {
            result = "digraph G{\n";
        } else {
            result = "graph G{\n";
        }
        result += toDot(uniqID, directed);
        result += "}";
        return result;
        
    }

    
    
    
    private Map<Integer,Map<Integer,Edge>> edges; // maps indices1[Int] to hashmaps, which map indices2[Int] to edges[Edge]
    //    private Map wordToIDs; // maps words to lists of IDs
    private Map<Integer,Node> nodes; // maps node IDs to nodes
    
    /** Add a node to the graph **/

    public void addNode(Node node) { // add a node to the graph
        Integer index = node.getIndex();
        nodes.put(index,node);
        // DEBUG
        // System.out.println("NODE INDEX: " + node.getIndex());
        // DEBUG
        // System.out.println("NODE WORD:  " + node.getWord());
        // DEBUG
        // System.out.println("NODE POS:   " + node.getPOS());
        // DEBUG
        // System.out.println("");
    }

    /** Get a node from the Graph. Throws an UnknownNodeException if an invalid index is being used. **/
    
    public Node getNode(Integer index) {
        if (!(nodes.containsKey(index))) {
            throw new UnknownNodeException("Unknown node with index "+index);
        } else {
        return (Node) nodes.get(index);
        }
    }

    /** Add an edge to the Graph. 
     *  The edge comparator decides whether to override an existing edge or not**/

    public void addEdge(Edge newEdge, Comparator<Edge> moreInformative) { // Add an Edge to the graph
    
        Integer from = newEdge.getFromNode().getIndex();
        Integer to = newEdge.getToNode().getIndex();
        Map<Integer,Edge> toMap;
        Map<Integer,Edge> fromMap;
    
    
        // DEBUG
        // System.out.println("-- EDGE FROM: " + from);
        // DEBUG
        // System.out.println("-- EDGE TO:   " + to);
        // DEBUG
        // System.out.println("");
    
    
    
        if (this.edges.containsKey(from)) {
            toMap = this.edges.get(from);
        } else {
            toMap = new HashMap<Integer,Edge>();
        }
        // if either this edge does not exist yet, or it is more informative, record it
        if ((!toMap.containsKey(to)) || (moreInformative.compare(newEdge,toMap.get(to)) > 0)) {     
            toMap.put(to,newEdge);
            this.edges.put(from,toMap);
        }
    

        if (this.edges.containsKey(to)) {
            fromMap = this.edges.get(to);
        } else {
            fromMap = new HashMap<Integer,Edge>();
        }
        Edge newEdgeInvert = newEdge.invert();
        // if either this edge does not exist yet, or it is more informative, record it
        if ((!fromMap.containsKey(from)) || (moreInformative.compare(newEdgeInvert, fromMap.get(from)) > 0)) {
            fromMap.put(from,newEdgeInvert);
            this.edges.put(to,fromMap);    
        } 
    }
    
    public void addEdge(Edge newEdge) { // Add an Edge to the graph
    
        // TODO: MAKE SURE THIS WORKS! ..I removed the comparator
    
        Integer from = newEdge.getFromNode().getIndex();
        Integer to = newEdge.getToNode().getIndex();
        Map<Integer,Edge> toMap;
        Map<Integer,Edge> fromMap;
    
    
    
        if (this.edges.containsKey(from)) {
            toMap = this.edges.get(from);
        } else {
            toMap = new HashMap<Integer,Edge>();
        }
        // if either this edge does not exist yet, or it is more informative, record it
        if ((!toMap.containsKey(to))) {     
            toMap.put(to, newEdge);
            this.edges.put(from, toMap);
        }
    

        if (this.edges.containsKey(to)) {
            fromMap = this.edges.get(to);
        } else {
            fromMap = new HashMap<Integer,Edge>();
        }
        Edge newEdgeInvert = newEdge.invert();
        // if either this edge does not exist yet, or it is more informative, record it
        if ((!fromMap.containsKey(from))) {
            fromMap.put(from,newEdgeInvert);
            this.edges.put(to,fromMap);    
        } 
    }
    
    
    
    /** Returns all edges for a given node **/
    public Map<Integer,Edge> getEdges(Integer vertice) {
        if (this.edges.containsKey(vertice)) {  
            return this.edges.get(vertice);
        } else {
            return new HashMap<Integer,Edge>();
        }
    }

    /** Returns the set of all node indices in the Graph. **/

    public Set<Integer> getAllNodeIndices() {
        return nodes.keySet();
    }

    /** Replace all occurrences of some index in the Graph by a new
     * index. This method does not check if the new ID is already in use. **/
    
    public void replaceBy(Integer oldID, Integer newID) { // traverse all edges and replace if necessary
        Map<Integer,Map<Integer,Edge>> newEdges = new HashMap<Integer,Map<Integer,Edge>>();
        for (Integer from: edges.keySet()) {    
            Map<Integer,Edge> thisEdges = edges.get(from);
            Map<Integer,Edge> newThisEdges = new HashMap<Integer,Edge>();
            for (Integer to: thisEdges.keySet()) {
                Edge thisEdge = thisEdges.get(to);
                Node fromNode = thisEdge.getFromNode();
                Node toNode = thisEdge.getToNode();
                String rel = ((Edge) thisEdges.get(to)).getRelation();
            
                if (from.equals(oldID)) {
                    fromNode = (Node) nodes.get(newID);
                }
                if (to.equals(oldID)) {
                    toNode = (Node) nodes.get(newID);
                }
                newThisEdges.put(to, new Edge(fromNode,toNode,rel));
            }
            newEdges.put(from,newThisEdges);
        }
        edges = newEdges; // overwrite old edges
        nodes.remove(oldID); // remove oldID from node list
    }

    /** Initialise an empty graph.
     **/
    
    public Graph() {   // Simplest way of turning a paragraph into a graph
        edges = new HashMap<Integer,Map<Integer,Edge>>(); 
        // wordToIDs = new HashMap();
        nodes = new HashMap<Integer,Node>();
    }

    /** Checks if there are no edges in the Graph. **/

    public boolean isEmpty() {
        return edges.isEmpty();
    }
    
    /** return a string representation of the graph. This has only been used for debugging purposes. **/
    public String toString() { // showing the graph...
        return "N: "+nodes.toString()+"\nE: "+edges.toString();
    }
}
