package de.ims.cl.dv.parsers;

import java.io.*;
import java.util.*;
import java.util.regex.*;

import de.ims.cl.dv.datastructures.*;
import de.ims.cl.dv.util.*;


class RelationLineError extends RuntimeException {
    private static final long serialVersionUID = 1;
    RelationLineError(String errormsg) {
        super(errormsg);
    }
}



public class CandCCorpus implements Corpus {    
    BufferedReader corpus;
    Logfile log;
    long    charcounter = 0; // count how much has been read from the corpus
    PosMapping   posmap = null;
    RelMapping   relmap = null;
    Pattern punctuation = Pattern.compile("^[^\\w]+$");
    
    Graph nextParse; // lookahead
    
    Map<String,String> relTrans = new HashMap<String,String>();
    
    private void increaseCounter(int inc) throws IOException { // increase charcounter and output
        long distance = 1000000;
        if ((charcounter/distance - (charcounter +=inc)/distance) < 0) { // printout every 10 mill lines
            log.println("Reading corpus byte "+(new Long(charcounter)).toString());
        } 
    }
    
    /**
       Returns the number of bytes read from the current corpus
    **/
    
    public long charcounter() {
        return charcounter;
    }
  

    private void initRelTrans(){
        // the transformation from original C&C relations to ours
        String transRules = "xmod -> mod | ncmod -> mod | cmod -> mod | ncsubj -> subj | xsubj -> subj | csubj -> subj | dobj -> obj | iobj -> obj | obj2 -> obj | xcomp -> xcomp | ccomp -> ccomp | det -> det | aux -> aux | conj -> conj";
        String orig, trans;
        String[] pairs;
        
        for (String pair: transRules.split(" \\| ")){
            pairs = pair.split(" -> ");
            orig = pairs[0];
            trans = pairs[1];
            relTrans.put(orig, trans);
        }
    }
    

    /**
       Returns the next parse graph
    **/
    
    private Edge edgeTransform(Edge edge){
        @SuppressWarnings("unused")
		Node  to   = edge.getToNode();
        Node  from = edge.getFromNode();
        String rel = edge.getRelation();
        if (rel.equals("subj") && from.getPOS().equals("det")) {
            from.setPOS("n");
        }
        return edge;
    }
    

    public Graph fetchNext() throws IOException {
    
        Graph pg = new Graph(); 
        List<PreliminaryEdge> edges = new LinkedList<PreliminaryEdge>(); // store preliminary edges
        // TODO: implement COORD resolution
        @SuppressWarnings("unused")
		String line,lemma,lemma_pos,rel;
        boolean error = false;
        PreliminaryEdge pe;
        Edge edge;
        

        line = corpus.readLine();
        
        while (true) {
            line = corpus.readLine();
            if (line == null) {
                return null;
            }
            
            increaseCounter(line.length());
            
            // line contains superfluous information:
            if (!line.startsWith("<") && !line.startsWith("(")) {
                continue;
            } else if (line.startsWith("(")) {
                // line contains edge information
                // ===============================
                // = RECOGNITION & TRANFORMATION =
                // ===============================
                
                try {
                    pe = handleEdgeLine(line);
                    if (pe == null) {
                        error = true;
                    }
                } catch (RelationLineError e) {
                    // e.printStackTrace();
                    error = true;
                    System.err.println(e);
                    continue;
                }
                
                
                // =====================
                // = INCORPORATE EDGES =
                // =====================
                
                edges.add(pe);
            } else if (line.startsWith("<c> ")) {
                // line contains POS info
                if (error) {
                    while (!line.equals("")){
                        // we have encountered a problem and must proceed to next graph
                        // which is separated via an empty line
                        line = corpus.readLine();
                    }
                    error = false;
                    return new Graph();
                } else {
                    // add nodes to graph
                    pg = (Graph) processPOSline(pg, line);
                    
                    // add edges to graph
                    for (PreliminaryEdge pre: edges) {
                        edge = pre.constructProperEdge(pg);
                        edge = edgeTransform(edge);
                        pg.addEdge(edge);
                    }
                    return pg;
                }
            }
            
        }  // end while
    }

    /** Read corpus from STDIN and write error output to <i>l</i>.  **/
    
       
    public CandCCorpus(Logfile l, Map<String,String> mappings) throws IOException{
        initRelTrans();
        this.log = l;
        this.corpus = new BufferedReader(new InputStreamReader(System.in));
               
        if (mappings.containsKey("posmapFile")) {
            String posmapFile = mappings.get("posmapFile");
            this.posmap = new PosMapping(posmapFile);
        }
        if (mappings.containsKey("relmapFile")) {
            String relmapFile = mappings.get("relmapFile");
            this.relmap = new RelMapping(relmapFile);
        }        
        nextParse = null;
        next();
    }


    public CandCCorpus(String filename, Logfile l, Map<String,String> mappings) throws FileNotFoundException {
        initRelTrans();
        this.corpus = new BufferedReader(new FileReader(filename));
        this.log = l;
        
        if (mappings.containsKey("posmapFile")) {
            String posmapFile = mappings.get("posmapFile");
            this.posmap = new PosMapping(posmapFile);
        }
        if (mappings.containsKey("relmapFile")) {
            String relmapFile = mappings.get("relmapFile");
            this.relmap = new RelMapping(relmapFile);
        }
        nextParse = null;
        next();
        
        
    }
    
    public CandCCorpus(String filename, Logfile l, String posmapFile, String relmapFile) throws FileNotFoundException {
        initRelTrans();
        this.corpus = new BufferedReader(new FileReader(filename));
        this.log = l;
        this.posmap = new PosMapping(posmapFile);
        this.relmap = new RelMapping(relmapFile);
        nextParse = null;
        next();
    }
    
        
    public Graph processPOSline(Graph g, String line){
        
        // KNOWLEDGE:
        // - the node information [from "(...)" lines] is contained in "<c>" line.
        //   >>> generally not vice versa! <<<
        //   e.g. punctuation and conjunctions are not included
        // - the "word_i" nodes are the token form "word" and the index "i" of the token
        //   in the "<c>" line.
        
        
        // line = "<c> The|the|DT|I-NP|O|NP[nb]/N common|common|JJ|I-NP|O|N/N characteristic|characteristic|NN|I-NP|O|N is|be|VBZ|I-VP|O|(S[dcl]\\NP)/NP the|the|DT|I-NP|O|NP[nb]/N happy|happy|JJ|I-NP|O|N/N acceptance|acceptance|NN|I-NP|O|N of|of|IN|I-PP|O|(NP\\NP)/NP whatever|whatever|WDT|I-NP|O|NP/(S[dcl]\\NP) comes|come|VBZ|I-VP|O|S[dcl]\\NP to|to|TO|I-VP|O|(S[to]\\NP)/(S[b]\\NP) mind|mind|VB|I-VP|O|(S[b]\\NP)/NP &hellip;|&hellip;|JJ|I-NP|O|N/N paintings|painting|NNS|I-NP|O|N are|be|VBP|I-VP|O|(S[dcl]\\NP)/NP a|a|DT|I-NP|O|NP[nb]/N stimulus|stimulus|NN|I-NP|O|N to|to|TO|I-PP|O|(NP\\NP)/NP pleasant|pleasant|JJ|I-NP|O|N/N experience|experience|NN|I-NP|O|N .|.|.|O|O|. &equo;|&equo;|NNP|I-NP|O|NP\\NP"
        String[] lines = line.split(" ");
        // lines[0] == "<c>"
        Integer lemmaIndex = 0;
        for (int i = 1; i < lines.length; i++) {
            String[] parts = lines[i].split("\\|");
            @SuppressWarnings("unused")
			String   token = parts[0];
            String   lemma = parts[1];
            String   pos   = parts[2];
            
            // check if lemma and type even fit (for sanity reasons)
            // if (!charsInCommon(token.toLowerCase(), lemma.toLowerCase())) {
            //     // DEBUG
            //     System.out.println("\n\n\n\n\nPROBLEM!\n\tTOKEN & LEMMA:\n\t'" + token + "' & '" + "'" + lemma + "'");
            // }
            
            if (this.posmap != null) {
                // DEBUG
                // System.err.println("looking in posmap");
                pos = this.posmap.lookup(pos);
            }
            
            g.addNode(new Node(lemma, pos, lemmaIndex));
            // Node node = buildNodeWithPos(...);
            
            lemmaIndex++;
        }
        
        // DEBUG
        // System.out.print("POS SENT: '");
        // DEBUG
        // System.out.print(posLine2Sent(line));
        // DEBUG
        // System.out.println("'");
        
        return g;
    }
    
    
    // private String[] transformRel(line)
    
    
    private boolean assertRelationFormat(String rel, String[] parts, Integer len){
        // assert (parts.length == len): ("Relation line for relation '" + rel + "' does have expected length " + len);
        // return true;
        if (parts.length - 1 != len) {
            System.err.println("Relation line for relation '" + rel + "' does not have expected length " + len);
            for (String part: parts){
                System.err.print(part + " ");
            }
            System.err.println();
            return false;
        } else {
            return true;
        }
    }
    
    private PreliminaryEdge handleEdgeLine(String line){
        
        
        String[] parts;
        String[] newparts;
        String rel;
        String trel;
        @SuppressWarnings("unused")
		String head, part, dep;
        Integer headIndex = -1;
        Integer depIndex = -1;
        
        
        parts = line.substring(1, line.length() - 1).split(" ");
        rel = parts[0];
        
        // throw away relation
        newparts = new String[parts.length - 1];
        
        for (int i=1; i < parts.length; i++){
            newparts[i-1] = parts[i];
        }
        
        if (parts.length < 2){
            // we have no real information
            return null;
        }
        if (relTrans.containsKey(rel)){
            trel = relTrans.get(rel);
        } else {
            trel = rel;
        }
        
        if (rel.equals("xmod") && assertRelationFormat(rel, parts, 3)){
            head = newparts[1].split("_")[0];
            headIndex = Integer.parseInt(newparts[1].split("_")[1]);
            dep  = newparts[2].split("_")[0];
            depIndex = Integer.parseInt(newparts[2].split("_")[1]);
        } else if (rel.equals("ncmod") && assertRelationFormat(rel, parts, 3)){
            head = newparts[1].split("_")[0];
            headIndex = Integer.parseInt(newparts[1].split("_")[1]);
            dep  = newparts[2].split("_")[0];
            depIndex = Integer.parseInt(newparts[2].split("_")[1]);
            trel = trel + "+" + newparts[0];
        } else if (rel.equals("cmod") && assertRelationFormat(rel, parts, 3)){
            head = newparts[1].split("_")[0];
            headIndex = Integer.parseInt(newparts[1].split("_")[1]);
            dep  = newparts[2].split("_")[0];
            depIndex = Integer.parseInt(newparts[2].split("_")[1]);
        } else if (rel.equals("ncsubj") && newparts[newparts.length - 1].equals("obj") && assertRelationFormat(rel, parts, 3)){
            head = newparts[0].split("_")[0];
            headIndex = Integer.parseInt(newparts[0].split("_")[1]);
            dep  = newparts[1].split("_")[0];
            depIndex = Integer.parseInt(newparts[1].split("_")[1]);
            trel = "obj";
        } else if (rel.equals("ncsubj") && assertRelationFormat(rel, parts, 3)){
            head = newparts[0].split("_")[0];
            headIndex = Integer.parseInt(newparts[0].split("_")[1]);
            dep  = newparts[1].split("_")[0];
            depIndex = Integer.parseInt(newparts[1].split("_")[1]);
        } else if (rel.equals("xsubj") && assertRelationFormat(rel, parts, 3)){
            head = newparts[0].split("_")[0];
            headIndex = Integer.parseInt(newparts[0].split("_")[1]);
            dep  = newparts[1].split("_")[0];
            depIndex = Integer.parseInt(newparts[1].split("_")[1]);
        } else if (rel.equals("csubj") && assertRelationFormat(rel, parts, 3)){
            head = newparts[0].split("_")[0];
            headIndex = Integer.parseInt(newparts[0].split("_")[1]);
            dep  = newparts[1].split("_")[0];
            depIndex = Integer.parseInt(newparts[1].split("_")[1]);
        } else if (rel.equals("dobj") && assertRelationFormat(rel, parts, 2)){
            head = newparts[0].split("_")[0];
            headIndex = Integer.parseInt(newparts[0].split("_")[1]);
            dep  = newparts[1].split("_")[0];
            depIndex = Integer.parseInt(newparts[1].split("_")[1]);
        } else if (rel.equals("iobj") && assertRelationFormat(rel, parts, 2)){
            head = newparts[0].split("_")[0];
            headIndex = Integer.parseInt(newparts[0].split("_")[1]);
            dep  = newparts[1].split("_")[0];
            depIndex = Integer.parseInt(newparts[1].split("_")[1]);
        } else if (rel.equals("obj2") && assertRelationFormat(rel, parts, 2)){
            head = newparts[0].split("_")[0];
            headIndex = Integer.parseInt(newparts[0].split("_")[1]);
            dep  = newparts[1].split("_")[0];
            depIndex = Integer.parseInt(newparts[1].split("_")[1]);
        } else if (rel.equals("xcomp") && assertRelationFormat(rel, parts, 3)){
            head = newparts[1].split("_")[0];
            headIndex = Integer.parseInt(newparts[1].split("_")[1]);
            dep  = newparts[2].split("_")[0];
            depIndex = Integer.parseInt(newparts[2].split("_")[1]);
            trel = trel + "+" + newparts[0];
        } else if (rel.equals("ccomp") && assertRelationFormat(rel, parts, 3)){
            head = newparts[1].split("_")[0];
            headIndex = Integer.parseInt(newparts[1].split("_")[1]);
            dep  = newparts[2].split("_")[0];
            depIndex = Integer.parseInt(newparts[2].split("_")[1]);
        } else if (rel.equals("aux") && assertRelationFormat(rel, parts, 2)){
            head = newparts[0].split("_")[0];
            headIndex = Integer.parseInt(newparts[0].split("_")[1]);
            dep  = newparts[1].split("_")[0];
            depIndex = Integer.parseInt(newparts[1].split("_")[1]);
        } else if (rel.equals("conj") && assertRelationFormat(rel, parts, 2)){
            head = newparts[0].split("_")[0];
            headIndex = Integer.parseInt(newparts[0].split("_")[1]);
            dep  = newparts[1].split("_")[0];
            depIndex = Integer.parseInt(newparts[1].split("_")[1]);
        } else if (rel.equals("det") && assertRelationFormat(rel, parts, 2)){
            head = newparts[0].split("_")[0];
            headIndex = Integer.parseInt(newparts[0].split("_")[1]);
            dep  = newparts[1].split("_")[0];
            depIndex = Integer.parseInt(newparts[1].split("_")[1]);
        } else {
            // return tuple(["unknown"] * 3);
            throw new RelationLineError("Relation line '" + line + "' has incorrect format.");
        }
        return new PreliminaryEdge(depIndex, headIndex, trel);
                
    }
    
    static public Boolean charsInCommon(String str1, String str2){
        for (int i = 0; i < str1.length(); i++) {
            for (int j = 0; j < str2.length(); j++) {
                if (str1.charAt(i) == str2.charAt(j)) {
                    return true;
                }
            }
        }
        return false;
    }
    
    
    public String posLine2Sent(String line){
        String res = "";
        String[] words = line.split(" ");
        for (int i = 1; i < words.length; i++) {
            String word = words[i].split("\\|")[0];
            res = res + word + " ";
        }
        res = res.substring(0, res.length()-1);
        return res;
    }
    // TODO: What is happening?
    // 1) Failure:
    //      Using path valuation function plain.
    //      Using context specification from contextspec_minimal.txt.
    //      [LLT] IO Error: java.io.IOException: Something went horribly wrong: totalnumber is zero.
    // 2)  vector files are empty

	  /**
     * return next parse graph - or null, if there isn't a next one.
     */
    public Graph next() {
        try {                
                Graph currentParse = nextParse;
                nextParse = fetchNext();
                return currentParse;
                
        } catch (IOException e) {
        throw new RuntimeException("IO Exception while reading corpus...");
        }
    }
    
    public boolean hasNext() {
        return (nextParse != null);
    }



	public void remove() {		
		  throw new UnsupportedOperationException();    
	}
    
}