/**
   class SemanticSpaceArray

   Copyright (C) 2002-4 Sebastian Pado
   
   This file is part of the DependencyVectors v2 distribution.
   
   DependencyVectors is free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later
   version.
   
   DependencyVectors is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public
   License along with DependencyVectors; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA

**/

package de.ims.cl.dv.space;

import java.util.*;
import java.io.*;
import java.text.*;

import de.ims.cl.dv.datastructures.FrequencyList;
import de.ims.cl.dv.util.Logfile;

class BEMatchError extends RuntimeException {
    private static final long serialVersionUID = 1;
    BEMatchError(String errormsg) {
	super(errormsg);
    }
}

class TargetNotFoundError extends RuntimeException {
    private static final long serialVersionUID = 1;
    TargetNotFoundError(String errormsg) {
	super(errormsg);
    }
}

/** This is a naive array-based implementation of the SemanticSpace
    interface.  It is rather fast, but requires much memory
    (<i>number(targets)*number(basis elements)*size(Double)</i>). **/

public class SemanticSpaceArray implements SemanticSpace { 
    
    final FrequencyList basisElements; // basisElements and frequency
    final FrequencyList targets;      // targets and frequency
    
    final List<String> be_indices = new ArrayList<String>(); // indexed list of bes
    final List<String> target_indices = new ArrayList<String>(); // indexed list of targets
    
    Logfile log;
    
    double[][] space;

  /** Increases the co-occurrence count for a given target and basis element combination.
     **/
      
    public void increaseCount(String target, String be, double value) {
	int row = target_indices.indexOf(target);
	int column = be_indices.indexOf(be);
	if ((row != -1) && (row < space.length) &&
	    (column != -1) && (column < space[0].length)) {
	    space[row][column] += value;
	}
    }

    /** Sets the co-occurrence count for a given target and basis
     * element combination to some number. **/

    public void setCount(String target, String be, double value) {
	int row = target_indices.indexOf(target);
	int column = be_indices.indexOf(be);
	if ((row != -1) && (row < space.length) &&
	    (column != -1) && (column < space[0].length)) {
	    space[row][column] = value;
	}
    }
    
    /** Performs a log-likelihood transformation for all the cells in
     * the space matrix. Requires the total number of co-occurrences
     * seen for the production of the given space to use the marginals
     * for efficient estimation.
     *
     * The present implementation is copied from Scott McDonald's vector.pl script.
    **/

    public void loglikelihood(double total) {
	
	for (int i=0; i < space.length; i++) {
	    String target = (String) target_indices.get(i);
	    double targetfreq = ((Double) targets.get(target)).doubleValue();
	    for (int j=0; j < space[i].length; j++) {
		String be = (String) be_indices.get(j);
		double befreq = ((Double) basisElements.get(be)).doubleValue();
		
		double a = space[i][j];
		double b = targetfreq - a;
		double c = befreq - a;
		double d = total - (b+c+a);

		if ((a > 0) && (b > 0) && (c>0) && (d>0)) {
		    space[i][j] = 
			2* (  a * Math.log(a) +
			      b * Math.log(b) +
			      c * Math.log(c) +
			      d * Math.log(d) - 
			      (  (a+b) * Math.log(a+b) +
				 (a+c) * Math.log(a+c) +
				 (b+d) * Math.log(b+d) +
		 				 (c+d) * Math.log(c+d) )  +
			      (a+b+c+d) * Math.log(a+b+c+d)   ) ;
		}
		else {
		    space[i][j] = 0;
		}
	    }
	}
    }

    public void pmi(double total) {
    	for (int i=0; i < space.length; i++) {
    		String target = (String) target_indices.get(i);
    		double targetfreq = ((Double) targets.get(target)).doubleValue();
    		for (int j=0; j < space[i].length; j++) {
    			String be = (String) be_indices.get(j);
    			double befreq = ((Double) basisElements.get(be)).doubleValue();
    			double coocc_freq = space[i][j];
    			if (coocc_freq > 0) {
    				space[i][j] = (coocc_freq * total)/(targetfreq*befreq);
    			}
    		}
    	}
    }
    
	public void locallogpmi(double total) {
		// TODO Auto-generated method stub
		for (int i=0; i< space.length; i++) {
			String target = (String) target_indices.get(i);
			double targetfreq = ((Double) targets.get(target).doubleValue());
			for (int j=0; j< space[i].length; j++) {
				String be = (String) be_indices.get(j);
				double befreq = ((Double) basisElements.get(be)).doubleValue();
				double coocc_freq = space[i][j];
				if (coocc_freq > 0) {
					space[i][j] = Math.log(coocc_freq) * Math.log((coocc_freq * total)/(targetfreq*befreq));
				}
			}
		}
	}
    
    public boolean containsTarget(String t) {
	return targets.containsKey(t); 
    }

    public boolean containsBE(String b) {
	return basisElements.containsKey(b); 
    }

    /** Returns the vector representing the target as a double array. **/
    public double[] getVector(String target) {
	return space[target_indices.indexOf(target)];
    }
    
    /** Returns a string representation of the SemanticSpace. The
      format follows the old McDonald file format. First line: list of
      basis elements (tab-separated). All following lines contain
      first a target, then a colon, and then space-separated double
      values.  Both targets and basis elements are ordered naturally
      (lexicographically). Notice that the resulting string object can
      become quite large due to the representation of floating point
      values by up to 16 characters each. It should probably be
      possible (but is not implemented) to specify the number of
      decimal places.  **/

    public String toString() { 
	
	DecimalFormat df = new DecimalFormat();
	df.setMaximumFractionDigits(4);
	df.setGroupingUsed(false);	
	StringBuffer returnstring = new StringBuffer();
	Iterator<String> beit = be_indices.iterator(); // first line: bes, tab-separated
 	while (beit.hasNext()) {
	    String next = (String) beit.next();
	    returnstring = returnstring.append(next);
	    returnstring = returnstring.append((beit.hasNext())?"\t":"\n");
	}
	Iterator<String> targetit = target_indices.iterator();
	//	System.err.println("# Targets: "+target_indices.size());
	while (targetit.hasNext()) {
	    String target = (String) targetit.next();
	    //	    try {
	    //	log.println("target "+target+" length of returnstring "+returnstring.length());
	    //} catch (IOException e) {
	    //}
  	    returnstring = returnstring.append(target);
	    returnstring = returnstring.append(": ");
	    double[] vector = getVector(target);
	    //	    System.err.println(target+" "+vector.length);
	    for (int i=0; i<vector.length; i++) {
		if (vector[i] == 0.0) {
		    returnstring = returnstring.append("0"); 
		} else {
		    returnstring = returnstring.append(df.format(vector[i]));
		}
		returnstring = returnstring.append((i==vector.length-1)?"\n":" ");
	    }
	}	
	return returnstring.toString();
    }
    

    /** Fill a semantic space with the data from a string representation of the space in a file. **/
    
    public void fillFromFile(String matrixfilename) throws IOException {
	
	BufferedReader file = new BufferedReader(new FileReader(matrixfilename));
	String line;	
	boolean is_firstLine = true;
	boolean hasWarned = false; 
	while (!((line = file.readLine()) == null)) {
	    if (is_firstLine) {
		String[] tabSplit = line.split("\t");
		if (tabSplit.length > 1) { // ok, first line contains bes
		    if (tabSplit.length >= be_indices.size()) { // check we have enough dimensions in the vector
			if (tabSplit.length > be_indices.size() && (!(hasWarned))) {
			    log.println("Warning: using only "+be_indices.size()+" of "+tabSplit.length+" dimensions.");
			    hasWarned = true;
			}
			for (int col=0; col < be_indices.size(); col++) {
			    // System.err.println(tabSplit[col]+" [sp] \tvs\t [be] "+be_indices.get(col));
			    
			    if (!(tabSplit[col].equals(be_indices.get(col)))) {
					throw new BEMatchError("BEs "+tabSplit[col]+" (space) and "+be_indices.get(col)+" (BE list) do not match!");
//					System.err.println("BEs "+tabSplit[col]+" (space) and "+be_indices.get(col)+" (BE list) do not match!");
			    } 
			} 
		    }
		    else {
			throw new BEMatchError("Too many BEs ("+be_indices.size()+") compared with matrix ("+tabSplit.length+")");
		    }
		    is_firstLine = false;
		    continue; 
		}  // LEFT: do check on targets as well!
	    }
	    String[] spaceSplit = line.split(" "); // normalise order to targets
	    if (spaceSplit[0].length() == 0) {
		System.err.println("[SemanticSpace] Ignoring line "+line);
		continue;
	    }
	    if (!(spaceSplit[0].substring(spaceSplit[0].length()-1).equals(":"))) {
		throw new TargetNotFoundError("No colon after target in line beginning with "+spaceSplit[0]);
	    }
	    
	    String target = spaceSplit[0].substring(0,spaceSplit[0].length()-1); // remove colon
	    int row = target_indices.indexOf(target);
	    if (row == -1) {
		throw new TargetNotFoundError("Cannot find target "+target+" in target list!");
	    }
	    for (int col=1; col < be_indices.size(); col++) {
		// System.out.println("row "+(row-1)+" col "+(col-1));
		space[row][col-1] = Double.parseDouble(spaceSplit[col]);
	    }
	}
    }
    
    /** Initializes a new (empty) SemanticSpace from a FrequencyList
     * of targets and a FrequencyList of basis elements. **/

    public SemanticSpaceArray(FrequencyList tws, FrequencyList bes, Logfile log) throws IOException {
	this.basisElements = bes;
	this.targets = tws;
	this.log = log;
	
	Iterator<Map.Entry<String,Double>> it = targets.entriesByFrequency().iterator(); 
	while (it.hasNext()) {
	    target_indices.add(it.next().getKey());
	}
	it = basisElements.entriesByFrequency().iterator();
	while (it.hasNext()) {	
	    be_indices.add(it.next().getKey());
	}
	       
	log.println("Allocating memory for new semantic space"); 
	log.println("Have "+tws.size()+" targets and "+bes.size()+" BEs");
	space = new double[tws.size()][bes.size()];

	log.println("Allocation successful");

	for (int i=0;i<tws.size();i++) {
	    for (int j=0;j<bes.size();j++) {
		space[i][j] = 0;
	    }
	}
    }

}

