/**
   class SemanticSpaceHash

   Copyright (C) 2002-4 Sebastian Pado
   
   This file is part of the DependencyVectors v2 distribution.
   
   DependencyVectors is free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later
   version.
   
   DependencyVectors is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public
   License along with DependencyVectors; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA

**/

package de.ims.cl.dv.space;

import java.util.*;
import java.text.*;

import de.ims.cl.dv.datastructures.FrequencyList;
import de.ims.cl.dv.util.Logfile;

/** This is a naive hash-based implementation of the SemanticSpace
    interface. Since it only needs the space of filled cells, and
    since most semantic spaces are very sparse, this implementation
    usually saves much memory over the SemanticSpaceArray
    implementation. However, each access involves two hashmap
    operations; therefore, it is slower . **/


public class SemanticSpaceHash implements SemanticSpace { 
                                       
    final FrequencyList basisElements; // basisElements and frequency
    final FrequencyList targets;      // targets and frequency
    Logfile log;
    
    Map<String,Map<String,Double>> space;
    
    
  /** Increases the co-occurrence count for a given target and basis element combination.
     **/
  
    public void increaseCount(String target, String be, double value) {
	Map<String,Double> vector = space.get(target);
	if (vector == null) {
	    System.err.println("Error: could not find vector for target "+target);
	    System.exit(0); }
	if (vector.containsKey(be)) {
	    double newValue = value + ((Double) vector.get(be)).doubleValue(); 
	    vector.put(be,new Double(newValue));
	    space.put(target,vector);
	} else {
	    vector.put(be, new Double(value));
	}
    }
    
    /** Determines if the target t is represented in the semantic space. **/
    
    public boolean containsTarget(String t) {
	return targets.containsKey(t); 
    }

    public boolean containsBE(String b) {
	return targets.containsKey(b); 
    }
    
    /** Returns a string representation of the semantic space. The
     * representation follows the format specified for
     * SemanticSpaceArray: The format follows the old McDonald file
     * format. First line: list of basis elements (tab-separated). All
     * following lines contain first a target, then a colon, and then
     * space-separated double values.  Both targets and basis elements
     * are ordered naturally (lexicographically). Notice that the
     * resulting string object can become quite large due to the
     * representation of floating point values by up to 16 characters
     * each. It should probably be possible (but is not implemented)
     * to specify the number of decimal places.  **/
    
    public String toString() { 

	DecimalFormat df = new DecimalFormat();
	df.setMaximumFractionDigits(4);
	df.setGroupingUsed(false);
	
	StringBuffer returnstring = new StringBuffer();
	List<Map.Entry<String,Double>> besSorted = basisElements.entriesByFrequency();

	// first line, print bes, tab-separated
	
	for (Map.Entry<String,Double> be: besSorted) {
		 String next = be.getKey();
		 returnstring = returnstring.append(next+"\t");
	}
	returnstring.setCharAt(returnstring.length()-1, '\n');
	
	
	Iterator<Map.Entry<String,Double>> tait = targets.entriesByFrequency().iterator(); 
	while (tait.hasNext()) {
	    String target = tait.next().getKey();
	    // log.println("target "+target+" length of returnstring "+returnstring.length());
	    //	    log.println("target "+target);
	    Map<String,Double> vector =  space.get(target);
	    if (vector == null) {
		System.err.println("Error: could not find vector for target "+target);
		System.exit(0); }
	    StringBuffer vectorstring = new StringBuffer(target).append(": ");

	    for (Map.Entry<String,Double> be: besSorted) {
	    	String beString = be.getKey();
			Object val = vector.get(beString);
			if (val != null) {
			    vectorstring = vectorstring.append(df.format(val));
			} else {
			    vectorstring = vectorstring.append("0");
			}
			vectorstring = vectorstring.append(" ");
	    }
	    vectorstring.setCharAt(vectorstring.length()-1,'\n');
	    
	    returnstring = returnstring.append(vectorstring);
	}
	return returnstring.toString();
    }

    /** Initialises a new (empty) SemanticSpace from a FrequencyList
     * of targets and a FrequencyList of basis elements. **/
    
    public SemanticSpaceHash(FrequencyList tws, FrequencyList bes, Logfile log) {
	this.basisElements = bes;
	this.targets = tws;
	this.log = log;
	
	space = new HashMap<String,Map<String,Double>>();
	
	Iterator<Map.Entry<String,Double>> tait = targets.entriesByFrequency().iterator(); 
	while (tait.hasNext()) {
	    String target = tait.next().getKey();
	    space.put(target,new HashMap<String,Double>(bes.size()));
	}
    }
}

