/**
   class LLT (executable)

   Copyright (C) 2002-4 Sebastian Pado
   
   This file is part of the DependencyVectors v2 distribution.
   
   DependencyVectors is free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later
   version.
   
   DependencyVectors is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public
   License along with DependencyVectors; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA

**/

package de.ims.cl.dv.space.transform;

import java.util.*;
import java.io.*; 

import de.ims.cl.dv.datastructures.ContextSpec;
import de.ims.cl.dv.datastructures.FrequencyList;
import de.ims.cl.dv.space.SemanticSpaceArray;
import de.ims.cl.dv.util.DVCommandLineMap;
import de.ims.cl.dv.util.DVIOHelpers;
import de.ims.cl.dv.util.Logfile;
import de.ims.cl.dv.util.Parameters;

/** <b>executable</b> This class wraps the lexical association function transformations.**/

public abstract class Transformation {

    static final String[] basisOptions = {"spec","help","targetdir","log"};
    static final String[] allValuations = Parameters.allValuations;
    static final String[] allCSFiles = Parameters.allCSFiles;
    String[][] targetcountsfile,becountsfile,totalcountsfile,matrixfile,outfile;
    String path;
    Logfile log;
    String[] valuation;
    ContextSpec[] cs;
    
    protected abstract void showHelp();
    
    /**
     * Constructor.
     * @param clm
     * @throws IOException
     */
    Transformation(DVCommandLineMap clm) throws IOException {

	if (clm.exists("help")) {
	    showHelp();
	}
	
	// prepare input/ output directory	
	
	path = clm.determineTargetDirectory();
	if (path == null) {
	    System.err.println("\"--targetdir\" command line argument missing. \n ");
	    showHelp();
	}
	
	// create logfile;

	log = clm.createLogfile(path+"LLT");
	
	// determine path valuation function(s)
	
	valuation = clm.determineValueFunctions(allValuations,log);
	
	// initialise context specification(s)

	cs = clm.determineContextSpecs(allCSFiles,log);
	
	// initialise filenames

	targetcountsfile = new String[cs.length][valuation.length];
	becountsfile = new String[cs.length][valuation.length];
	totalcountsfile = new String[cs.length][valuation.length];
	matrixfile = new String[cs.length][valuation.length];
	outfile = new String[cs.length][valuation.length];
	
	for (int i=0; i<cs.length; i++) {
	    for (int j=0; j<valuation.length; j++){
		String affix = DVIOHelpers.affix(cs[i].name(),valuation[j]);
		targetcountsfile[i][j] = path + affix +".targets";
		becountsfile[i][j] =  path + affix + ".bes";
		totalcountsfile[i][j] = path + affix + ".total";
		matrixfile[i][j] = path + affix + ".vectors";
		outfile[i][j] =  path + affix + ".ll.vectors";
		DVIOHelpers.testLocation(outfile[i][j]); // test if the out location is writeable
	    }
	}
    }
    
    /** open total number file and read in totalnumber **/
    private final double readTotalFromFile(String numberfilename) throws IOException { 
	
	double totalnumber;
	String line;

	BufferedReader file = new BufferedReader(new FileReader(numberfilename));
	if (!((line = file.readLine()) == null)) {
	    try {
		totalnumber = Double.parseDouble(line);
	    }
	    catch (NumberFormatException e) {
		throw new IOException("Invalid total number: "+line);
	    }
	} else {
	    throw new IOException("Total Number file is empty.");
	}
	file.close();
	
	if (totalnumber == 0.0) {
	    throw new IOException("Something went horribly wrong: totalnumber is zero.");
	}
	return totalnumber;
    }
    
    /** this method is called from within transformSpaces() and calls the actual transformation. **/
    protected abstract void callTransformation(SemanticSpaceArray s, double totalnumber);

    /** Transform and write to file **/
	protected final void transformSpaces() throws IOException {
	
	for (int i=0; i<cs.length; i++) {
	    for (int j=0; j<valuation.length; j++){	    
		
		FrequencyList basisElements = new FrequencyList(becountsfile[i][j], log, true); // read basis elements file
		FrequencyList targets = new FrequencyList(targetcountsfile[i][j], log, true);	 // read basis elements file
		double totalnumber = readTotalFromFile(totalcountsfile[i][j]);
		SemanticSpaceArray semanticSpace = new SemanticSpaceArray(targets,basisElements,log);
		semanticSpace.fillFromFile(matrixfile[i][j]);
	
		// compute transformation		
		callTransformation(semanticSpace, totalnumber);
		// write to file
		DVIOHelpers.writeToFile(outfile[i][j],semanticSpace.toString()); 
	    }
	}
    }
    
     /** This function is called when running the program with <i>java
     * LLT /< arguments />. The program relies on receiving the right
     * command line arguments and will exit if they are not
     * supplied. See help output and documentation for details about
     * arguments. </i>
     **/
   
    static public DVCommandLineMap doInitialization(String[] argarray) {
    	List<String> recognisedOptions = new ArrayList<String>();
   	    recognisedOptions.addAll(Arrays.asList(basisOptions));
   	    recognisedOptions.addAll(Arrays.asList(Parameters.allValuations));
   	    DVCommandLineMap arguments = new DVCommandLineMap(argarray,recognisedOptions);
   	    return arguments;
    }
    
}
