/**
   class DVCommandLineMap

   Copyright (C) 2002-4 Sebastian Pado
   
   This file is part of the DependencyVectors v2 distribution.
   
   DependencyVectors is free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later
   version.
   
   DependencyVectors is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public
   License along with DependencyVectors; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA

**/

package de.ims.cl.dv.util;

import java.util.*;
import java.io.*;
import java.net.*;

import de.ims.cl.dv.ExtractBasisElements;
import de.ims.cl.dv.datastructures.ContextSpec;
import de.ims.cl.dv.parsers.*;

/** This class provides a number of methods to interpret the arguments
 * given on the command line. It is specifically adapted to the needs
 * of the DependencyVectors package. **/

public class DVCommandLineMap extends CommandLineMap {

   Encoding e;

    public String validPOS() {
    if (this.exists("pos")) {
        return e.encode(this.get("pos"));
    } else {
        return null;
    }   
    }

   public String getBasisMapping(Logfile log) throws IOException{
     if (this.exists("basisMapping")) {
       return this.get("basisMapping");       
     } else {
       System.err.println("Error: Basis Mapping (lex or gramlex)  must be specified!");
       log.println("Error: Basis Mapping (lex or gramlex) must be specified!");
       System.exit(0); 
        return ""; // we never get here.
     }
   }

    /** determines target directory. Returns null if target directory not specified. **/
    
    public String determineTargetDirectory() {
    if (this.exists("targetdir")) {
        
        if (! new File(this.get("targetdir")).isDirectory()) {
        System.err.println("The argument to targetdir is not a directory. Exiting. \n ");
        System.exit(0);
        }
        
        if (this.get("targetdir").endsWith(File.separator))
        { return this.get("targetdir"); }
        else {
        return this.get("targetdir")+"/"; }
    }   else {
        return null;
    }
    }

    /** creates a logfile according to the current command line **/

    public Logfile createLogfile(String path) throws IOException {
    Logfile log;
    if (this.exists("log")) {
        log = new Logfile(path+this.get("log"));
        log.println("Logfile is called "+path+this.get("log")+".");
    } else {
        log = new Logfile(path+"log.txt");
        log.println("Logfile is called "+path+"log.txt.");
    }
    return log;
    }

    
    
    
    
    
    /** determines the corpus specified by the current command line **/

    public Corpus determineCorpus(Logfile log) throws IOException {
    Corpus corpus;
    //  
    // corpus = new MiniparCorpus(log); // read from System.in
           
    	if (this.get("ctype").equals("candc") ){

    		// for CnC, use the standard POS mapping/REL mapping by default
    		
    		Map<String,String>  mappings  = new HashMap<String, String>();

			System.err.println("Reading corpus in CandC format");
			log.println("Reading corpus in CandC format");
          
            if (this.exists("posmap")){
            	 mappings.put("posmapFile",this.get("posmap"));
            } else {            	 
            	mappings.put("posmapFile","posmap_cnc_default.txt");
            }

            if (this.exists("relmap")) {
            	mappings.put("relmapFile",this.get("relmap"));
            } else {
            	mappings.put("relmapFile","relmap_cnc_default.txt");
            }
            
            if (this.exists("corpus")) {            
            	  corpus = new CandCCorpus(this.get("corpus"), log, mappings);
            } else {
                  corpus = new CandCCorpus(log, mappings);
                          
            }
    	  } else if(this.get("ctype").equals("freeling")){
              
              Map<String,String>  mappings  = new HashMap<String, String>();

              System.err.println("Reading corpus in Freeling dependencies format");
              log.println("Reading corpus in Freeling dependencies format");
        
          if (this.exists("posmap")){
               mappings.put("posmapFile",this.get("posmap"));
          } else {                     
              mappings.put("posmapFile","posmap_freeling_default.txt");
          }

          if (this.exists("relmap")) {
              mappings.put("relmapFile",this.get("relmap"));
          } else {
              mappings.put("relmapFile","relmap_freeling_default.txt");
          }
          
          if (this.exists("corpus")) {            
                corpus = new FreelingCorpus(this.get("corpus"), log, mappings);
          } else {
                corpus = new FreelingCorpus(log, mappings);
                        
          }

    	} else {
    		
    		if (this.get("ctype").equals("minipar")) {
     			System.err.println("Reading corpus in MINIPAR format");
     			log.println("Reading corpus in MINIPAR format");
     		    		} else {	
     		    System.err.println("Defaulting to MINIPAR format");
     		    log.println("Defaulting to MINIPAR format");
    		}
    		
            if (this.exists("corpus")) {            
            	corpus = new MiniparCorpus(this.get("corpus"), log);
            } else {
            	corpus = new MiniparCorpus(log);                    
            }
    	}            
    	return corpus;
    }

    /** determines the context specifications specified by the current command line **/
    
    public ContextSpec[] determineContextSpecs(String[] allCSFiles, Logfile log) throws IOException {
    
    ContextSpec[] cs;
    
    
    URLClassLoader classLoader = (URLClassLoader)ExtractBasisElements.class.getClassLoader(); // get access to class path
    
    if (this.exists("spec")) {
        allCSFiles = new String[1];
        allCSFiles[0] = this.get("spec");
    } 
    
    cs = new ContextSpec[allCSFiles.length]; // check if context specs exist (and exit gracefully if not)
    for (int l = 0; l < allCSFiles.length; l++) {
    	  String thisCSFileName = "contextspec_"+allCSFiles[l]+".txt";
    	  log.println("Using context specification from "+thisCSFileName+".");
          System.err.println("Using context specification from "+thisCSFileName+".");
          URL resourcePath = classLoader.getResource(thisCSFileName);
          if (resourcePath == null) {
          System.err.println("Error: Context specification "+thisCSFileName+" not found.");
          log.println("Error: Context specification "+thisCSFileName+" not found.");
          System.exit(0); }       
        cs[l] = new ContextSpec(classLoader.getResource(thisCSFileName).getFile());
//        System.err.println(classLoader.getResource(thisCSFileName).getFile());
    }
    
    return cs;
    }    

    /** determines the value functions specified by the current command line **/ 
    
    public String[] determineValueFunctions(String[] allValuations, Logfile log) throws IOException {
    
    String[] valuation;
    
    if (this.exists("plain")) {
        System.err.println("Using path valuation function plain.");
        log.println("Using path valuation function plain.");
        valuation = new String[1];
        valuation[0] = "plain"; }
    else if (this.exists("length")) {
        System.err.println("Using path valuation function with penalized length.");
        log.println("Using path valuation function with penalized length.");
        valuation = new String[1];
        valuation[0] = "length"; }
    else if (this.exists("oblique")) {
        System.err.println("Using obliqueness hierarchy path valuation function."); 
        log.println("Using obliqueness hierarchy path valuation function."); 
        valuation = new String[1];
        valuation[0] = "oblique"; }
    else if (this.exists("oblength")) {
        System.err.println("Using oblique-and-length path valuation function.");
        log.println("Using oblique-and-length path valuation function.");
        valuation = new String[1];
        valuation[0] = "oblength"; }
//    else if (this.exists("obl_steep")) {
//        System.err.println("Using steep oblique path valuation function.");
//        log.println("Using steep oblique path valuation function.");
//        valuation = new String[1];
//        valuation[0] = "obl_steep"; }
//    else if (this.exists("obl_flat")) {
//        System.err.println("Using flat oblique path valuation function.");
//        log.println("Using flat oblique path valuation function.");
//        valuation = new String[1];
//        valuation[0] = "obl_flat"; }
//    else if (this.exists("obl_inv")) {
//        System.err.println("Using inverse oblique path valuation function.");
//        log.println("Using inverse oblique path valuation function.");
//        valuation = new String[1];
//        valuation[0] = "obl_inv"; }
    else {
        System.err.println("Using ALL path valuation functions.");
        log.println("Using ALL path valuation functions.");
        valuation = allValuations;
    }
    return valuation;
    }    

    public DVCommandLineMap(String[] args, List<String> knownArgs) {
    super(args,knownArgs);
    e = new MiniparEncoding();
    }

}
