/**
   class Parameters

   Copyright (C) 2002-4 Sebastian Pado
   
   This file is part of the DependencyVectors v2 distribution.
   
   DependencyVectors is free software; you can redistribute it
   and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later
   version.
   
   DependencyVectors is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
   PURPOSE.  See the GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public
   License along with DependencyVectors; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
   MA 02111-1307 USA

**/


package de.ims.cl.dv.util;

import java.util.*;
import java.io.*;

import de.ims.cl.dv.datastructures.Edge;
import de.ims.cl.dv.datastructures.Path;

/** The class Parameters contains functions that parametrize the extraction of Basis Elements and Semantic Spaces. 
    They can be changed by the user. 
**/

// Computes a partition of paths by mapping them onto a label representing this partition
interface BasisMapping {
  String basisMapping(Path path);
}

// Map each path onto the last word plus the last relation of the path
class GramLexBasisMapping implements BasisMapping {
  public String basisMapping(Path path) {
    Edge lastEdge = (Edge) path.get(path.size()-1);
    return lastEdge.getToNode().getWord()+"+"+lastEdge.getRelation();
  }
}

  // Map each path onto the sequence of its edge labels - no lexical information whatsoever
class GramBasisMapping implements BasisMapping {
	public String basisMapping(Path path) {
		StringBuffer sequence = new StringBuffer();
		for (Edge e: path) {
			sequence.append(e.getRelation()+"_");
		}
		sequence.deleteCharAt(sequence.length()-1);
		return sequence.toString();
	}
}
 
// Map each path onto the last word
class LexBasisMapping implements BasisMapping {
  public String basisMapping(Path path) {
    Edge lastEdge = (Edge) path.get(path.size()-1);
    return lastEdge.getToNode().getWord();
  }
  
}


class PlainValuation {
  static public double evaluatePath(Path path) {
    return 1;
  }
}
class LengthValuation {
  static public double evaluatePath(Path path) {
    if (path.size() == 1)
      return 3;
        else if (path.size() ==2)
      return 2;
        else return 1;
  }
}
class ObliqueValuation {
  static public double evaluatePath(Path path) {
    int rv;
    HashSet<String> rels = path.relSet();
    if (rels.contains("subj")) {    // subjects
  rv = 5; }
    else if (rels.contains("obj")) { // objects
  rv = 4; }
    else if (rels.contains("pcomp-n")) {  // oblique objects
  rv = 3; }
    else if (rels.contains("gen")) { // genitives
  rv = 2; }
    else {         // modifiers and other stuff
  rv = 1; }
    return rv;
  }
}


public class Parameters {
  
  BasisMapping basisMapping;
  
    /** lists all context specifications **/

    public static String[] allCSFiles = {"minimal", // search in classpath (lib/contextspecs)
				  "medium",
				  "rich",
				  "wide",
	  			  "all3v",
				  "all3n"};
    
    /** pretty print path value function names for help **/
    
    public static String prettyPrintSpecs() {
	StringBuffer result = new StringBuffer();
	if (allCSFiles.length > 0) {
	    result.append(allCSFiles[0]);
	    for (int i=1; i<allCSFiles.length; i++) {
		result.append(", ");
		result.append(allCSFiles[i]);
	    }
	}
	return result.toString();
    }
    
    /** list all path value functions **/

    public static String[] allValuations = {"plain","length","oblique"};

    /** pretty print path value function names for help **/

    public static String prettyPrintVals() {
	StringBuffer result = new StringBuffer();
	if (allValuations.length > 0) {
	    result.append("--"+allValuations[0]);
	    for (int i=1; i<allValuations.length; i++) {
		result.append(" | ");
		result.append("--"+allValuations[i]);
	    }
	}
	return result.toString();
    }

 
      public String basisMapping(Path path) {
        return basisMapping.basisMapping(path);
      }
    
    
    /** pathValue implements the Path Value Function parameter.
	It provides a counting scheme for paths
	in the form of a function path x value function -> int
	that returns the value of a path for a given valuation function.
	Currently implemented path value functions are: plain, oblique, length, oblength.

       By adding new counting schemes, the user can specify his own Path Value Function. <br>
    **/
    
  public double pathValue(Path path, String valuation) {
    if (valuation.equals("plain")) {
      return PlainValuation.evaluatePath(path);
    } else if (valuation.equals("length")) {
      return LengthValuation.evaluatePath(path);
    } else if (valuation.equals("oblique")) {
      return ObliqueValuation.evaluatePath(path);
    } else {
      throw new RuntimeException("Error in Parameters: unknown path valuation function "+valuation);
    }
  }
    
  public Parameters(String mapping, Logfile log) throws IOException {    
    if (mapping.equals("lex")) {
      basisMapping = new LexBasisMapping();
      System.err.println("Choosing LexBasisMapping function.");
      log.println("Choosing LexBasisMapping function.");
    } else if (mapping.equals("gramlex")) {
        basisMapping = new GramLexBasisMapping();
        System.err.println("Choosing GramLexBasisMapping function.");
        log.println("Choosing GramLexBasisMapping function.");
    } else if (mapping.equals("gram")) {
        basisMapping = new GramBasisMapping();
        System.err.println("Choosing GramBasisMapping function.");
        log.println("Choosing GramBasisMapping function.");
    }  else {
      throw new RuntimeException("Error in Parameters: unknown basis mapping function "+basisMapping);
    }   
  }
}
